// neoPLC IMU library
// for BMX055 9-axis inertial measurement unit
// Deft Dynamics 2016
// based on code by Kris Wiener

#include "arduino.h"
#include "neoPLC_IMU.h"
#include <Wire.h>


/**************************************************
* Function name	: void neoPLC_IMU()
*    returns		: void
* Created by		: Austin Gurley
* Date created		: 8/27/16
* Description		: constructor for neoPLC_IMU class 
* Notes			: constructor
**************************************************/
neoPLC_IMU::neoPLC_IMU()
{
}

/**************************************************
* Function name	: bool begin(int acc_range, int acc_bw, int gyro_range, int gyro_bw, int mag_odr)
*    returns		: logical 1 for successful intialization, 0 for failure
*    acc_range		: accelerometer full scale range (see table in header file)
*    acc_bw   		: accelerometer bandwidth (see table in header file)
*    gyro_range		: gyroscope full scale range (see table in header file)
*    gyro_bw		: gyroscope bandwidth (see table in header file)
*    mag_odr		: magnetometer output data rate (see table in header file)
* Created by		: Austin Gurley
* Date created		: 8/27/16
* Description		: connects to BMX055 and calls private funtions to initialize all 3 sensors.  
* Notes			: N/A
**************************************************/
bool neoPLC_IMU::begin(int acc_range, int acc_bw, int gyro_range, int gyro_bw, int mag_odr)
{
  Wire.begin();
  uint8_t a=0;
  uint8_t b=0;
  uint8_t c=0;
  a = readByte(BMX055_ACC_ADDR, BMX055_ACC_WHOAMI) == 0xFA;  // Read acc WHO_AM_I register and check value
  b = readByte(BMX055_GYRO_ADDR, BMX055_GYRO_WHOAMI) == 0x0F;  // Read gyro WHO_AM_I register and check value
     writeByte(BMX055_MAG_ADDR, BMX055_MAG_PWR_CNTL1, 0x01); // wake up magnetometer
     delay(10);
  c = readByte(BMX055_MAG_ADDR, BMX055_MAG_WHOAMI) == 0x32 ;  // Read mag WHO_AM_I register and check value

  if ( a & b & c ) // if all 3 sensors connect, complete the setup and return success flag
  {
    setupAcc(acc_range, acc_bw);
    setupGyro(gyro_range, gyro_bw);
    setupMag(mag_odr);
    return 1;
  }
  else             // if any of 3 sensors doesn't connect, return failure flag
  {
    return 0;
  }
}

/**************************************************
* Function name	: void readAccel(float *ax, float *ay, float *az)
*    returns		: using pointers
*    *ax			: pointer to variable to hold X accel
*    *ay			: pointer to variable to hold Y accel
*    *az			: pointer to variable to hold Z accel
* Created by		: Austin Gurley
* Date created		: 8/27/16
* Description		: reads BMX055 accelerometer, saves data at input pointers if new data is available 
* Notes			: N/A
**************************************************/
void neoPLC_IMU::readAccel(float *ax, float *ay, float *az)
{
  uint8_t rawData[6];
  int16_t raw_ax;
  int16_t raw_ay;
  int16_t raw_az;
  readBytes(BMX055_ACC_ADDR, BMX055_ACC_D_X_LSB, 6, &rawData[0]);      
  if((rawData[0] & 0x01) && (rawData[2] & 0x01) && (rawData[4] & 0x01)) // only update if NEW data is present
  { 
    raw_ax = (int16_t) (((int16_t)rawData[1] << 8) | rawData[0]) >> 4;  // Turn the MSB and LSB into a signed 12-bit value
    raw_ay = (int16_t) (((int16_t)rawData[3] << 8) | rawData[2]) >> 4;  
    raw_az = (int16_t) (((int16_t)rawData[5] << 8) | rawData[4]) >> 4; 
    *ax = (float)raw_ax*_acc_res;
    *ay = (float)raw_ay*_acc_res;
    *az = (float)raw_az*_acc_res;
  }
}

/**************************************************
* Function name	: void readGyro(float *gx, float *gy, float *gz)
*    returns		: using pointers
*    *gx			: pointer to variable to hold X rotation rate
*    *gy			: pointer to variable to hold Y rotation rate
*    *gz			: pointer to variable to hold Z rotation rate
* Created by		: Austin Gurley
* Date created		: 8/27/16
* Description		: reads BMX055 gyroscope, saves data at input pointers
* Notes			: N/A
**************************************************/
void neoPLC_IMU::readGyro(float *gx, float *gy, float *gz)
{
  uint8_t rawData[6];
  int16_t raw_gx;
  int16_t raw_gy;
  int16_t raw_gz;
  readBytes(BMX055_GYRO_ADDR, BMX055_GYRO_RATE_X_LSB, 6, &rawData[0]);  // Read the six raw data registers sequentially into data array
    raw_gx = (int16_t) (((int16_t)rawData[1] << 8) | rawData[0]);   // Turn the MSB and LSB into a signed 16-bit value
    raw_gy = (int16_t) (((int16_t)rawData[3] << 8) | rawData[2]);  
    raw_gz = (int16_t) (((int16_t)rawData[5] << 8) | rawData[4]); 
    *gx = (float)raw_gx*_gyro_res;
    *gy = (float)raw_gy*_gyro_res;
    *gz = (float)raw_gz*_gyro_res;
}

/**************************************************
* Function name	: void readMag(float *mx, float *my, float *mz)
*    returns		: using pointers
*    *mx			: pointer to variable to hold X magnetic flux
*    *my			: pointer to variable to hold Y magnetic flux
*    *mz			: pointer to variable to hold Z magnetic flux
* Created by		: Austin Gurley
* Date created		: 8/27/16
* Description		: reads BMX055 magnetometer, saves data at input pointers if new data is available 
* Notes			: N/A
**************************************************/
void neoPLC_IMU::readMag(float *mx, float *my, float *mz)
{
  int16_t mdata_x = 0, mdata_y = 0, mdata_z = 0, temp = 0;
  uint16_t data_r = 0;
  uint8_t rawData[8];  // x/y/z hall magnetic field data, and Hall resistance data
  int16_t raw_mx=0;
  int16_t raw_my=0;
  int16_t raw_mz=0;
  
  readBytes(BMX055_MAG_ADDR, BMX055_MAG_XOUT_LSB, 8, &rawData[0]);  // Read the eight raw data registers sequentially into data array
    if(rawData[6] & 0x01) // data ready
    {
      mdata_x = (int16_t) (((int16_t)rawData[1] << 8) | rawData[0]) >> 3;  // 13-bit signed integer for x-axis field
      mdata_y = (int16_t) (((int16_t)rawData[3] << 8) | rawData[2]) >> 3;  // 13-bit signed integer for y-axis field
      mdata_z = (int16_t) (((int16_t)rawData[5] << 8) | rawData[4]) >> 1;  // 15-bit signed integer for z-axis field
      data_r = (uint16_t) (((uint16_t)rawData[7] << 8) | rawData[6]) >> 2;  // 14-bit unsigned integer for Hall resistance
  
     // calculate temperature compensated 16-bit magnetic fields
     temp = ((int16_t)(((uint16_t)((((int32_t)_dig_xyz1) << 14)/(data_r != 0 ? data_r : _dig_xyz1))) - ((uint16_t)0x4000)));
     raw_mx = ((int16_t)((((int32_t)mdata_x) *
          ((((((((int32_t)_dig_xy2) * ((((int32_t)temp) * ((int32_t)temp)) >> 7)) +
             (((int32_t)temp) * ((int32_t)(((int16_t)_dig_xy1) << 7)))) >> 9) +
           ((int32_t)0x100000)) * ((int32_t)(((int16_t)_dig_x2) + ((int16_t)0xA0)))) >> 12)) >> 13)) + (((int16_t)_dig_x1) << 3);
     raw_my = ((int16_t)((((int32_t)mdata_y) *
          ((((((((int32_t)_dig_xy2) * ((((int32_t)temp) * ((int32_t)temp)) >> 7)) + 
             (((int32_t)temp) * ((int32_t)(((int16_t)_dig_xy1) << 7)))) >> 9) +
                 ((int32_t)0x100000)) * ((int32_t)(((int16_t)_dig_y2) + ((int16_t)0xA0)))) >> 12)) >> 13)) +
        (((int16_t)_dig_y1) << 3);
     raw_mz = (((((int32_t)(mdata_z - _dig_z4)) << 15) - ((((int32_t)_dig_z3) * ((int32_t)(((int16_t)data_r) -
           ((int16_t)_dig_xyz1))))>>2))/(_dig_z2 + ((int16_t)(((((int32_t)_dig_z1) * ((((int16_t)data_r) << 1)))+(1<<15))>>16))));
    *mx = (float)raw_mx*_mag_res;
    *my = (float)raw_my*_mag_res;
    *mz = (float)raw_mz*_mag_res;
   }
}

/**************************************************
* Function name  : void readTemp()
*    returns    : die temperature in deg C
* Created by    : Austin Gurley
* Date created    : 8/27/16
* Description   : reads BMX055 accelerometer die temperature
* Notes     : N/A
**************************************************/
float neoPLC_IMU::readTemp()
{
  int c =  readByte(BMX055_ACC_ADDR, BMX055_ACC_D_TEMP);  // Read the raw data register (8-bit = 1-byte)
  int16_t t = ((int16_t)((int16_t)c << 8)) >> 8;
  float temp = ((float)t/1.42+23); // read accelerometer temp and convert to deg C
  return temp;  // Turn the byte into a signed 8-bit integer
}

/**************************************************
* Function name	: void setupAcc(int acc_range, int acc_bw)
*    returns		: void
*    acc_range		: desired accelerometer range (must be valid value, see table in header)
*    acc_bw 		: desired accelerometer bandwidth (must be valid value, see table in header)
* Created by		: Austin Gurley
* Date created		: 8/27/16
* Description		: sets the useful parameter registers of the BMX055 accelerometer
* Notes			: private
**************************************************/
void neoPLC_IMU::setupAcc(int acc_range, int acc_bw)
{
  uint8_t Ascale = 0x03;
  switch (acc_range)
  {
    case 2:
      _acc_res = 2.0/2048.0; // 2 G
      Ascale = 0x03; break;
    case 4:
      _acc_res = 4.0/2048.0; // 4 G
      Ascale = 0x05; break;
    case 8:
      _acc_res = 8.0/2048.0; // 8 G
      Ascale = 0x08; break;
    case 16:
      _acc_res = 16.0/2048.0; // 16 G
      Ascale = 0x0C; break;
    default:
      _acc_res = 2.0/2048.0; // default: 2 G
      Ascale = 0x03; break;
  }

  uint8_t Abw;
  switch (acc_bw)
  {
    case 8:
      Abw = 0x00; break; // 7.81 Hz,  64 ms update time
    case 16:
      Abw = 0x01; break; // 15.63 Hz, 32 ms update time
    case 32:
      Abw = 0x02; break; // 31.25 Hz, 16 ms update time
    case 64:
      Abw = 0x03; break; // 62.5  Hz,  8 ms update time
    case 125:
      Abw = 0x04; break; // 125   Hz,  4 ms update time
    case 250:
      Abw = 0x05; break; // 250   Hz,  2 ms update time
    case 500:
      Abw = 0x06; break; // 500   Hz,  1 ms update time
    case 1000:
      Abw = 0x07; break; // 1000  Hz,  0.5 ms update time
    default:
      Abw = 0x04; break; // default to: 125  Hz,  4 ms update time
  }
  
   writeByte(BMX055_ACC_ADDR,  BMX055_ACC_BGW_SOFTRESET, 0xB6);  // reset accelerometer
   delay(1000); // Wait for all registers to reset 
   writeByte(BMX055_ACC_ADDR, BMX055_ACC_PMU_RANGE, Ascale & 0x0F);    // Set accelerometer full range
   writeByte(BMX055_ACC_ADDR, BMX055_ACC_PMU_BW, (0x08 | Abw) & 0x0F); // Set accelerometer bandwidth
   writeByte(BMX055_ACC_ADDR, BMX055_ACC_D_HBW, 0x00);                  // Use filtered data

    _dig_x1 = readByte(BMX055_ACC_ADDR, BMM050_DIG_X1);
    _dig_x2 = readByte(BMX055_ACC_ADDR, BMM050_DIG_X2);
    _dig_y1 = readByte(BMX055_ACC_ADDR, BMM050_DIG_Y1);
    _dig_y2 = readByte(BMX055_ACC_ADDR, BMM050_DIG_Y2);
    _dig_xy1 = readByte(BMX055_ACC_ADDR, BMM050_DIG_XY1);
    _dig_xy2 = readByte(BMX055_ACC_ADDR, BMM050_DIG_XY2);
}

/**************************************************
* Function name	: void setupGyro(int gyro_range, int gyro_bw)
*    returns		: void
*    gyro_range		: desired gyroscope range (must be valid value, see table in header)
*    gyro_bw 		: desired gyroscope bandwidth (must be valid value, see table in header)
* Created by		: Austin Gurley
* Date created		: 8/27/16
* Description		: sets the useful parameter registers of the BMX055 gyroscope
* Notes			: private
**************************************************/
void neoPLC_IMU::setupGyro(int gyro_range, int gyro_bw)
{
  uint8_t Gscale = 0x03;
  switch (gyro_range)
  {
    case 125:
      _gyro_res = 124.87/32768.0; 
      Gscale = 0x04; break;
    case 250:
      _gyro_res = 249.75/32768.0; 
      Gscale = 0x03; break;
    case 500:
      _gyro_res = 499.5/32768.0; 
      Gscale = 0x02; break;
    case 1000:
      _gyro_res = 999.0/32768.0; 
      Gscale = 0x01; break;
    case 2000:
      _gyro_res = 1998.0/32768.0; 
      Gscale = 0x00; break;
    default:
      _gyro_res = 249.75/32768.0; 
      Gscale = 0x03; break;
  }

  uint8_t Gbw;
  switch (gyro_bw)
  {
    case 523:
      Gbw = 0x00; break; // 2000 Hz ODR, 523 Hz BW
    case 230:
      Gbw = 0x01; break; // 2000 Hz ODR, 230 Hz BW
    case 116:
      Gbw = 0x02; break; // 1000 Hz ODR, 116 Hz BW
    case 47:
      Gbw = 0x03; break; //  400 Hz ODR,  47 Hz BW
    case 23:
      Gbw = 0x04; break; //  200 Hz ODR,  23 Hz BW
    case 12:
      Gbw = 0x05; break; //  100 Hz ODR,  12 Hz BW
    case 64:
      Gbw = 0x06; break; //  200 Hz ODR,  64 Hz BW
    case 32:
      Gbw = 0x07; break; //  100 Hz ODR,  32 Hz BW
    default:
      Gbw = 0x02; break; // default to: 1000 Hz ODR, 116 Hz BW
  }
  
  writeByte(BMX055_GYRO_ADDR, BMX055_GYRO_RANGE, Gscale);  // set GYRO FS range
  writeByte(BMX055_GYRO_ADDR, BMX055_GYRO_BW, Gbw);     // set GYRO ODR and Bandwidth

}

/**************************************************
* Function name	: void setupMag(int mag_odr)
*    returns		: void
*    mag_odr		: desired magnetometer output data rate (must be valid value, see table in header)
* Created by		: Austin Gurley
* Date created		: 8/27/16
* Description		: sets the useful parameter registers of the BMX055 magnetometer. defaults to enhanced mode
* Notes			: private
**************************************************/
void neoPLC_IMU::setupMag(int mag_odr)
{
    uint8_t Mbw;
  switch (mag_odr)
  {
    case 10:
      Mbw = 0x00; break; // 10 Hz ODR
    case 2:
      Mbw = 0x01; break; //  2 Hz ODR
    case 6:
      Mbw = 0x02; break; //  6 Hz ODR
    case 8:
      Mbw = 0x03; break; //  8 Hz ODR
    case 15:
      Mbw = 0x04; break; // 15 Hz ODR
    case 20:
      Mbw = 0x05; break; // 20 Hz ODR
    case 25:
      Mbw = 0x06; break; // 25 Hz ODR
    case 30:
      Mbw = 0x07; break; // 30 Hz ODR
    default:
      Mbw = 0x07; break; // default to: 30 Hz ODR
  }
  
  writeByte(BMX055_MAG_ADDR, BMX055_MAG_PWR_CNTL1, 0x82);  // Softreset magnetometer, ends up in sleep mode
  delay(100);
  writeByte(BMX055_MAG_ADDR, BMX055_MAG_PWR_CNTL1, 0x01); // Wake up magnetometer
  delay(100);
  writeByte(BMX055_MAG_ADDR, BMX055_MAG_PWR_CNTL2, Mbw << 3); // Normal mode
  
         // Low-power
       //   writeByte(BMX055_MAG_ADDR, BMX055_MAG_REP_XY, 0x01);  // 3 repetitions (oversampling)
       //   writeByte(BMX055_MAG_ADDR, BMX055_MAG_REP_Z,  0x02);  // 3 repetitions (oversampling)
          // Regular
       //   writeByte(BMX055_MAG_ADDR, BMX055_MAG_REP_XY, 0x04);  //  9 repetitions (oversampling)
       //   writeByte(BMX055_MAG_ADDR, BMX055_MAG_REP_Z,  0x16);  // 15 repetitions (oversampling)
          // Enhanced Regular
          writeByte(BMX055_MAG_ADDR, BMX055_MAG_REP_XY, 0x07);  // 15 repetitions (oversampling)
          writeByte(BMX055_MAG_ADDR, BMX055_MAG_REP_Z,  0x22);  // 27 repetitions (oversampling)
          // High Accuracy
        //  writeByte(BMX055_MAG_ADDR, BMX055_MAG_REP_XY, 0x17);  // 47 repetitions (oversampling)
        //  writeByte(BMX055_MAG_ADDR, BMX055_MAG_REP_Z,  0x51);  // 83 repetitions (oversampling)

      
        uint8_t rawData[2];  //placeholder for 2-byte trim data
          readBytes(BMX055_MAG_ADDR, BMM050_DIG_Z1_LSB, 2, &rawData[0]);   
        _dig_z1 = (uint16_t) (((uint16_t)rawData[1] << 8) | rawData[0]);  
          readBytes(BMX055_MAG_ADDR, BMM050_DIG_Z2_LSB, 2, &rawData[0]);   
        _dig_z2 = (int16_t) (((int16_t)rawData[1] << 8) | rawData[0]);  
          readBytes(BMX055_MAG_ADDR, BMM050_DIG_Z3_LSB, 2, &rawData[0]);   
        _dig_z3 = (int16_t) (((int16_t)rawData[1] << 8) | rawData[0]);  
          readBytes(BMX055_MAG_ADDR, BMM050_DIG_Z4_LSB, 2, &rawData[0]);   
        _dig_z4 = (int16_t) (((int16_t)rawData[1] << 8) | rawData[0]);  
         readBytes(BMX055_MAG_ADDR, BMM050_DIG_XYZ1_LSB, 2, &rawData[0]);   
        _dig_xyz1 = (uint16_t) (((uint16_t)rawData[1] << 8) | rawData[0]);  

        _mag_res =  1.0/1.6;
}


/**************************************************
* Function name	: void writeByte(uint8_t ic_address , uint8_t reg_address, uint8_t byte_to_send)
*    returns		: void
*    ic_address		: the slave address of the i2c device
*    reg_address	: the register subaddress
*    byte_to_send	: the byte to send
* Created by		: Austin Gurley
* Date created		: 8/27/16
* Description		: writes a single byte to an i2c register 
* Notes			: private
**************************************************/
void neoPLC_IMU::writeByte(uint8_t ic_address , uint8_t reg_address, uint8_t byte_to_send)
{
  Wire.beginTransmission(ic_address);
  Wire.write(reg_address);
  Wire.write(byte_to_send);
  Wire.endTransmission();
}

/**************************************************
* Function name	: uint8_t readByte(uint8_t ic_address , uint8_t reg_address)
*    returns		: the byte that is read
*    ic_address		: the slave address of the i2c device
*    reg_address	: the register subaddress
* Created by		: Austin Gurley
* Date created		: 8/27/16
* Description		: read a single byte in an i2c register. no pause between address writing and request
* Notes			: private
**************************************************/
uint8_t neoPLC_IMU::readByte(uint8_t ic_address, uint8_t reg_address)
{
  uint8_t message;
  Wire.beginTransmission(ic_address);
  Wire.write(reg_address);
  Wire.endTransmission(0);        
  Wire.requestFrom(ic_address, (size_t) 1);
  message = Wire.read();
  return message;
}

/**************************************************
* Function name	: void readBytes(uint8_t ic_address , uint8_t reg_address, uint8_t num_bytes, uint8_t *output_location)
*    returns			: using pointers
*    ic_address			: the slave address of the i2c device
*    reg_address		: the register subaddress
*    num_bytes			: number of bytes expected to recieve
*    *output_location	: pointer to the desired output starting point
* Created by			: Austin Gurley
* Date created			: 8/27/16
* Description			: read multiple bytes in an i2c register. no pause between address writing and request
* Notes			: private
**************************************************/
void neoPLC_IMU::readBytes(uint8_t ic_address, uint8_t reg_address, uint8_t num_bytes, uint8_t *output_location)
{
  uint8_t i = 0;
  Wire.beginTransmission(ic_address);
  Wire.write(reg_address);
  Wire.endTransmission(0);
  Wire.requestFrom(ic_address, (size_t) num_bytes);
  while (Wire.available()) 
        {
          output_location[i++] = Wire.read();
        }  
}



